# -*- coding: utf-8 -*-
"""
Tests of the functions xml_setter_xpaths
"""
import os
from lxml import etree
import pytest

FILE_PATH = os.path.dirname(os.path.abspath(__file__))
TEST_INPXML_PATH = os.path.join(FILE_PATH, 'files/fleur/Max-R5/FePt_film_SSFT_LO/files/inp2.xml')


def test_xml_create_tag_schema_dict(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_xpaths import xml_create_tag_schema_dict

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    los_before = eval_xpath(root, '/fleurInput/atomSpecies/species/lo')

    assert [node.getparent().attrib['name'] for node in los_before] == ['Fe-1', 'Fe-1', 'Pt-1']
    assert [node.attrib.items() for node in los_before] == [[('type', 'SCLO'), ('l', '0'), ('n', '3'), ('eDeriv', '0')],
                                                            [('type', 'SCLO'), ('l', '1'), ('n', '3'), ('eDeriv', '0')],
                                                            [('type', 'SCLO'), ('l', '1'), ('n', '5'), ('eDeriv', '0')]]

    xml_create_tag_schema_dict(xmltree, schema_dict, '/fleurInput/atomSpecies/species',
                               '/fleurInput/atomSpecies/species', 'lo')

    los_after = eval_xpath(root, '/fleurInput/atomSpecies/species/lo')

    assert [node.getparent().attrib['name'] for node in los_after] == ['Fe-1', 'Fe-1', 'Fe-1', 'Pt-1', 'Pt-1']
    assert [node.attrib.items() for node in los_after] == [[],
                                                           [('type', 'SCLO'), ('l', '0'), ('n', '3'), ('eDeriv', '0')],
                                                           [('type', 'SCLO'), ('l', '1'), ('n', '3'), ('eDeriv', '0')],
                                                           [],
                                                           [('type', 'SCLO'), ('l', '1'), ('n', '5'), ('eDeriv', '0')]]


def test_xml_create_tag_schema_dict_differing_xpaths(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_xpaths import xml_create_tag_schema_dict

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    los_before = eval_xpath(root, '/fleurInput/atomSpecies/species/lo')

    assert [node.getparent().attrib['name'] for node in los_before] == ['Fe-1', 'Fe-1', 'Pt-1']
    assert [node.attrib.items() for node in los_before] == [[('type', 'SCLO'), ('l', '0'), ('n', '3'), ('eDeriv', '0')],
                                                            [('type', 'SCLO'), ('l', '1'), ('n', '3'), ('eDeriv', '0')],
                                                            [('type', 'SCLO'), ('l', '1'), ('n', '5'), ('eDeriv', '0')]]

    xml_create_tag_schema_dict(xmltree, schema_dict, "/fleurInput/atomSpecies/species[@name='Fe-1']",
                               '/fleurInput/atomSpecies/species', 'lo')

    los_after = eval_xpath(root, '/fleurInput/atomSpecies/species/lo')

    assert [node.getparent().attrib['name'] for node in los_after] == ['Fe-1', 'Fe-1', 'Fe-1', 'Pt-1']
    assert [node.attrib.items() for node in los_after] == [[],
                                                           [('type', 'SCLO'), ('l', '0'), ('n', '3'), ('eDeriv', '0')],
                                                           [('type', 'SCLO'), ('l', '1'), ('n', '3'), ('eDeriv', '0')],
                                                           [('type', 'SCLO'), ('l', '1'), ('n', '5'), ('eDeriv', '0')]]


def test_xml_create_tag_schema_dict_element(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_xpaths import xml_create_tag_schema_dict

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    new_element = etree.Element('ldaU')
    new_element.attrib['test_attrib'] = 'test'

    assert len(eval_xpath(root, '/fleurInput/atomSpecies/species/ldaU', list_return=True)) == 0

    xml_create_tag_schema_dict(xmltree, schema_dict, '/fleurInput/atomSpecies/species',
                               '/fleurInput/atomSpecies/species', new_element)

    nodes = eval_xpath(root, '/fleurInput/atomSpecies/species/ldaU', list_return=True)

    assert [node.getparent().attrib['name'] for node in nodes] == ['Fe-1', 'Pt-1']
    assert [node.attrib.items() for node in nodes] == [[('test_attrib', 'test')], [('test_attrib', 'test')]]


def test_xml_create_tag_schema_dict_create_parents(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_xpaths import xml_create_tag_schema_dict

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    with pytest.raises(ValueError, match="Could not create tag 'addArg' because atleast one subtag is missing."):
        xml_create_tag_schema_dict(xmltree, schema_dict, "/fleurInput/atomSpecies/species[@name='Fe-1']/ldaHIA",
                                   '/fleurInput/atomSpecies/species/ldaHIA', 'addArg')

    xml_create_tag_schema_dict(xmltree,
                               schema_dict,
                               "/fleurInput/atomSpecies/species[@name='Fe-1']/ldaHIA",
                               '/fleurInput/atomSpecies/species/ldaHIA',
                               'addArg',
                               create_parents=True)

    assert len(eval_xpath(root, '/fleurInput/atomSpecies/species/ldaHIA/addArg', list_return=True)) == 1


def test_eval_xpath_create_existing(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_xpaths import eval_xpath_create

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    assert len(eval_xpath(root, '/fleurInput/atomSpecies/species/lo')) == 3

    nodes = eval_xpath_create(xmltree, schema_dict, '/fleurInput/atomSpecies/species/lo',
                              '/fleurInput/atomSpecies/species/lo')

    assert len(nodes) == 3


def test_eval_xpath_create_non_existing(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_xpaths import eval_xpath_create

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    assert len(eval_xpath(root, '/fleurInput/atomSpecies/species/ldaU')) == 0

    nodes = eval_xpath_create(xmltree, schema_dict, '/fleurInput/atomSpecies/species/ldaU',
                              '/fleurInput/atomSpecies/species/ldaU')

    assert len(nodes) == 2
    assert [node.getparent().attrib['name'] for node in nodes] == ['Fe-1', 'Pt-1']


def test_eval_xpath_create_differing_xpaths(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_xpaths import eval_xpath_create

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    assert len(eval_xpath(root, '/fleurInput/atomSpecies/species/ldaU')) == 0

    nodes = eval_xpath_create(xmltree,
                              schema_dict,
                              "/fleurInput/atomSpecies/species[@name='Fe-1']/ldaU",
                              '/fleurInput/atomSpecies/species/ldaU',
                              list_return=True)

    assert len(nodes) == 1
    assert [node.getparent().attrib['name'] for node in nodes] == ['Fe-1']


def test_eval_xpath_create_create_parents(load_inpxml):

    from masci_tools.util.xml.xml_setters_xpaths import eval_xpath_create

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)

    with pytest.raises(ValueError, match="Could not create tag 'addArg' because atleast one subtag is missing."):
        eval_xpath_create(xmltree, schema_dict, "/fleurInput/atomSpecies/species[@name='Fe-1']/ldaHIA/addArg",
                          '/fleurInput/atomSpecies/species/ldaHIA/addArg')

    nodes = eval_xpath_create(xmltree,
                              schema_dict,
                              "/fleurInput/atomSpecies/species[@name='Fe-1']/ldaHIA/addArg",
                              '/fleurInput/atomSpecies/species/ldaHIA/addArg',
                              create_parents=True,
                              list_return=True)

    assert len(nodes) == 1


TEST_ATTRIB_NAME = ['Kmax', 'kmax', 'KMAX']
TEST_VALUES = ['9.000000', 5.321, 'Pi/4.0']
TEST_RESULTS = ['9.000000', '5.3210000000', 'Pi/4.0']


@pytest.mark.parametrize('attribname, attribvalue, result', zip(TEST_ATTRIB_NAME, TEST_VALUES, TEST_RESULTS))
def test_xml_set_attrib_value(load_inpxml, attribname, attribvalue, result):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_xpaths import xml_set_attrib_value

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    xml_set_attrib_value(xmltree, schema_dict, '/fleurInput/calculationSetup/cutoffs',
                         '/fleurInput/calculationSetup/cutoffs', attribname, attribvalue)

    assert str(eval_xpath(root, '/fleurInput/calculationSetup/cutoffs/@Kmax')) == result


#The integer argument is converted to float since the arguemnt radius is float_expression (either float or str)
TEST_ATTRIB_RESULTS = [['test', 'test'], ['test', 'test2'], ['test', '2214.0000000000'], ['test', '2.20000000'],
                       ['2.20000000', 'test']]
TEST_ATTRIBV = ['test', ['test', 'test2'], ['test', 2214], 'test', ['test']]
TEST_OCCURENCES = [None, None, None, 0, [-1]]


@pytest.mark.parametrize('attribv, expected_result,occurrences', zip(TEST_ATTRIBV, TEST_ATTRIB_RESULTS,
                                                                     TEST_OCCURENCES))
def test_xml_set_attrib_value_all(load_inpxml, attribv, expected_result, occurrences):
    """
    Test of the functionality of xml_set_attrib_value_no_create with multiple occurrences
    of the sttribute and different values for occurrences
    """
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_xpaths import xml_set_attrib_value

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    assert eval_xpath(root, '/fleurInput/atomSpecies/species/mtSphere/@radius') == ['2.20000000', '2.20000000']

    xmltree = xml_set_attrib_value(xmltree,
                                   schema_dict,
                                   '/fleurInput/atomSpecies/species/mtSphere',
                                   '/fleurInput/atomSpecies/species/mtSphere',
                                   'radius',
                                   attribv,
                                   occurrences=occurrences)

    assert eval_xpath(root, '/fleurInput/atomSpecies/species/mtSphere/@radius') == expected_result


def test_xml_set_attrib_value_differing_xpaths(load_inpxml):
    """
    Test of the functionality of xml_set_attrib_value_no_create with multiple occurrences
    of the attribute and different values for occurrences
    """
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_xpaths import xml_set_attrib_value

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    assert eval_xpath(root, '/fleurInput/atomSpecies/species/mtSphere/@radius') == ['2.20000000', '2.20000000']

    xmltree = xml_set_attrib_value(xmltree, schema_dict, "/fleurInput/atomSpecies/species[@name='Fe-1']/mtSphere",
                                   '/fleurInput/atomSpecies/species/mtSphere', 'radius', 'TEST')

    assert eval_xpath(root, '/fleurInput/atomSpecies/species/mtSphere/@radius') == ['TEST', '2.20000000']


def test_xml_set_attrib_value_create(load_inpxml):
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_xpaths import xml_set_attrib_value

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    with pytest.raises(ValueError, match="Could not set attribute 'key' on path "):
        xml_set_attrib_value(xmltree, schema_dict, "/fleurInput/atomSpecies/species[@name='Fe-1']/ldaHIA/addArg",
                             '/fleurInput/atomSpecies/species/ldaHIA/addArg', 'key', 'TEST')

    xml_set_attrib_value(xmltree,
                         schema_dict,
                         "/fleurInput/atomSpecies/species[@name='Fe-1']/ldaHIA/addArg",
                         '/fleurInput/atomSpecies/species/ldaHIA/addArg',
                         'key',
                         'TEST',
                         create=True)

    assert eval_xpath(root, '/fleurInput/atomSpecies/species/ldaHIA/addArg/@key') == 'TEST'


def test_xml_set_first_attrib_value(load_inpxml):
    """
    Test of the functionality of xml_set_attrib_value_no_create with multiple occurrences
    of the sttribute and different values for occurrences
    """
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_xpaths import xml_set_first_attrib_value

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    assert eval_xpath(root, '/fleurInput/atomSpecies/species/mtSphere/@radius') == ['2.20000000', '2.20000000']

    xmltree = xml_set_first_attrib_value(xmltree, schema_dict, '/fleurInput/atomSpecies/species/mtSphere',
                                         '/fleurInput/atomSpecies/species/mtSphere', 'radius', 'TEST')

    assert eval_xpath(root, '/fleurInput/atomSpecies/species/mtSphere/@radius') == ['TEST', '2.20000000']


def test_xml_set_first_attrib_value_create(load_inpxml):
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_xpaths import xml_set_first_attrib_value

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    with pytest.raises(ValueError, match="Could not set attribute 'key' on path "):
        xml_set_first_attrib_value(xmltree, schema_dict, "/fleurInput/atomSpecies/species[@name='Fe-1']/ldaHIA/addArg",
                                   '/fleurInput/atomSpecies/species/ldaHIA/addArg', 'key', 'TEST')

    xml_set_first_attrib_value(xmltree,
                               schema_dict,
                               "/fleurInput/atomSpecies/species[@name='Fe-1']/ldaHIA/addArg",
                               '/fleurInput/atomSpecies/species/ldaHIA/addArg',
                               'key',
                               'TEST',
                               create=True)

    assert eval_xpath(root, '/fleurInput/atomSpecies/species/ldaHIA/addArg/@key') == 'TEST'


def test_xml_set_text(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_xpaths import xml_set_text

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    xml_set_text(xmltree, schema_dict, '/fleurInput/comment', '/fleurInput/comment', 'TEST_COMMENT')

    assert str(eval_xpath(root, '/fleurInput/comment/text()')) == 'TEST_COMMENT'


TEST_TEXT_RESULTS = [
    [' 1.0000000000000  1.0000000000000  1.0000000000000', ' 1.0000000000000  1.0000000000000  1.0000000000000'],
    ['1.0 1.0 1.0', '1.0 1.0 1.0'],
    [' 1.0000000000000  1.0000000000000  1.0000000000000', ' 2.0000000000000  2.0000000000000  2.0000000000000'],
    ['   -0.250000     0.250000     0.000000', '-20 30 40']
]
TEST_TEXTS = [[1.0, 1.0, 1.0], ['1.0', '1.0', '1.0'], [[1.0, 1.0, 1.0], [2.0, 2.0, 2.0]], ['-20', '30', '40']]
TEST_TEXT_OCCURENCES = [None, None, None, [-1]]


@pytest.mark.parametrize('text, expected_result,occurrences', zip(TEST_TEXTS, TEST_TEXT_RESULTS, TEST_TEXT_OCCURENCES))
def test_xml_set_text_all(load_inpxml, text, expected_result, occurrences):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_xpaths import xml_set_text

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    kpoints_xpath = '/fleurInput/cell/bzIntegration/kPointLists/kPointList/kPoint'

    assert eval_xpath(root, f'{kpoints_xpath}/text()') == [
        '   -0.250000     0.250000     0.000000', '    0.250000     0.250000     0.000000'
    ]

    xmltree = xml_set_text(xmltree, schema_dict, kpoints_xpath, kpoints_xpath, text, occurrences=occurrences)

    assert eval_xpath(root, f'{kpoints_xpath}/text()') == expected_result


def test_xml_set_text_differing_xpaths(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_xpaths import xml_set_text

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    kpoints_xpath = '/fleurInput/cell/bzIntegration/kPointLists/kPointList/kPoint'

    assert eval_xpath(root, f'{kpoints_xpath}/text()') == [
        '   -0.250000     0.250000     0.000000', '    0.250000     0.250000     0.000000'
    ]

    xmltree = xml_set_text(xmltree, schema_dict, f'{kpoints_xpath}[2]', kpoints_xpath, ['1.0', '1.0', '1.0'])

    assert eval_xpath(root, f'{kpoints_xpath}/text()') == ['   -0.250000     0.250000     0.000000', '1.0 1.0 1.0']


def test_xml_set_text_create(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_xpaths import xml_set_text

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    with pytest.raises(ValueError, match='Could not set text on path'):
        xml_set_text(xmltree, schema_dict,
                     "/fleurInput/atomSpecies/species[@name='Fe-1']/torgueCalculation/greensfElements/s",
                     '/fleurInput/atomSpecies/species/torgueCalculation/greensfElements/s', [False, False, True, False])

    xml_set_text(xmltree,
                 schema_dict,
                 "/fleurInput/atomSpecies/species[@name='Fe-1']/torgueCalculation/greensfElements/s",
                 '/fleurInput/atomSpecies/species/torgueCalculation/greensfElements/s', [False, False, True, False],
                 create=True)

    assert eval_xpath(root, '/fleurInput/atomSpecies/species/torgueCalculation/greensfElements/s/text()') == 'F F T F'


def test_xml_set_first_text(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_xpaths import xml_set_first_text

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    kpoints_xpath = '/fleurInput/cell/bzIntegration/kPointLists/kPointList/kPoint'

    assert eval_xpath(root, f'{kpoints_xpath}/text()') == [
        '   -0.250000     0.250000     0.000000', '    0.250000     0.250000     0.000000'
    ]

    xmltree = xml_set_first_text(xmltree, schema_dict, kpoints_xpath, kpoints_xpath, [1.0, 1.0, 1.0])

    assert eval_xpath(root, f'{kpoints_xpath}/text()') == [
        ' 1.0000000000000  1.0000000000000  1.0000000000000', '    0.250000     0.250000     0.000000'
    ]


def test_xml_set_first_text_create(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_xpaths import xml_set_first_text

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    with pytest.raises(ValueError, match='Could not set text on path'):
        xml_set_first_text(xmltree, schema_dict,
                           "/fleurInput/atomSpecies/species[@name='Fe-1']/torgueCalculation/greensfElements/s",
                           '/fleurInput/atomSpecies/species/torgueCalculation/greensfElements/s',
                           [False, False, True, False])

    xml_set_first_text(xmltree,
                       schema_dict,
                       "/fleurInput/atomSpecies/species[@name='Fe-1']/torgueCalculation/greensfElements/s",
                       '/fleurInput/atomSpecies/species/torgueCalculation/greensfElements/s',
                       [False, False, True, False],
                       create=True)

    assert eval_xpath(root, '/fleurInput/atomSpecies/species/torgueCalculation/greensfElements/s/text()') == 'F F T F'


TEST_SHIFT_ATTRIB_NAME = ['Kmax', 'kmax', 'KMAX']
TEST_SHIFT_VALUES = ['9.000000', 5.321, '9.000000']
TEST_SHIFT_RESULTS = ['13.0000000000', '9.3210000000', '13.0000000000']


@pytest.mark.parametrize('attribname, shift_value, result',
                         zip(TEST_SHIFT_ATTRIB_NAME, TEST_SHIFT_VALUES, TEST_SHIFT_RESULTS))
def test_xml_add_number_to_attrib(load_inpxml, attribname, shift_value, result):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_xpaths import xml_add_number_to_attrib

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    xml_add_number_to_attrib(xmltree, schema_dict, '/fleurInput/calculationSetup/cutoffs',
                             '/fleurInput/calculationSetup/cutoffs', attribname, shift_value)

    assert str(eval_xpath(root, '/fleurInput/calculationSetup/cutoffs/@Kmax')) == result


TEST_SHIFT_VALUES_ALL = [1.0, 0.5, 3.0]
TEST_SHIFT_RESULTS_ALL_ABS = [['3.2000000000', '3.2000000000'], ['2.7000000000', '2.7000000000'],
                              ['5.2000000000', '2.20000000']]
TEST_SHIFT_RESULTS_ALL_REL = [['2.2000000000', '2.2000000000'], ['1.1000000000', '1.1000000000'],
                              ['6.6000000000', '2.20000000']]
TEST_SHIFT_OCCURRENCES = [None, None, 0]


@pytest.mark.parametrize('shift_value, result, occurrences',
                         zip(TEST_SHIFT_VALUES_ALL, TEST_SHIFT_RESULTS_ALL_ABS, TEST_SHIFT_OCCURRENCES))
def test_xml_add_number_to_attrib_all_abs(load_inpxml, shift_value, result, occurrences):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_xpaths import xml_add_number_to_attrib

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    xml_add_number_to_attrib(xmltree,
                             schema_dict,
                             '/fleurInput/atomSpecies/species/mtSphere',
                             '/fleurInput/atomSpecies/species/mtSphere',
                             'radius',
                             shift_value,
                             occurrences=occurrences)

    radius = eval_xpath(root, '/fleurInput/atomSpecies/species/mtSphere/@radius')

    assert [str(val) for val in radius] == result


@pytest.mark.parametrize('shift_value, result, occurrences',
                         zip(TEST_SHIFT_VALUES_ALL, TEST_SHIFT_RESULTS_ALL_REL, TEST_SHIFT_OCCURRENCES))
def test_xml_add_number_to_attrib_all_rel(load_inpxml, shift_value, result, occurrences):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_xpaths import xml_add_number_to_attrib

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    xml_add_number_to_attrib(xmltree,
                             schema_dict,
                             '/fleurInput/atomSpecies/species/mtSphere',
                             '/fleurInput/atomSpecies/species/mtSphere',
                             'radius',
                             shift_value,
                             occurrences=occurrences,
                             mode='rel')

    radius = eval_xpath(root, '/fleurInput/atomSpecies/species/mtSphere/@radius')

    assert [str(val) for val in radius] == result


def test_xml_add_number_to_attrib_differing_xpaths(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_xpaths import xml_add_number_to_attrib

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    xml_add_number_to_attrib(xmltree, schema_dict, "/fleurInput/atomSpecies/species[@name='Fe-1']/mtSphere",
                             '/fleurInput/atomSpecies/species/mtSphere', 'radius', 2.0)

    radius = eval_xpath(root, '/fleurInput/atomSpecies/species/mtSphere/@radius')

    assert [str(val) for val in radius] == ['4.2000000000', '2.20000000']


def test_xml_add_number_to_attrib_differing_int(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_xpaths import xml_add_number_to_attrib

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    xml_add_number_to_attrib(xmltree, schema_dict, '/fleurInput/calculationSetup/cutoffs',
                             '/fleurInput/calculationSetup/cutoffs', 'numbands', 100)

    numbands = eval_xpath(root, '/fleurInput/calculationSetup/cutoffs/@numbands')

    assert numbands == '100'

    with pytest.raises(ValueError, match='You are trying to write a float to an integer attribute'):
        xml_add_number_to_attrib(xmltree, schema_dict, '/fleurInput/calculationSetup/cutoffs',
                                 '/fleurInput/calculationSetup/cutoffs', 'numbands', 55.5)


def test_xml_add_number_to_first_attrib(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_xpaths import xml_add_number_to_first_attrib

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    xml_add_number_to_first_attrib(xmltree, schema_dict, '/fleurInput/atomSpecies/species/mtSphere',
                                   '/fleurInput/atomSpecies/species/mtSphere', 'radius', 2.0)

    radius = eval_xpath(root, '/fleurInput/atomSpecies/species/mtSphere/@radius')

    assert [str(val) for val in radius] == ['4.2000000000', '2.20000000']


def test_xml_set_simple_tag_single(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_xpaths import xml_set_simple_tag

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    changes = {'s': 2, 'p': 3, 'd': 4, 'f': 5}

    xml_set_simple_tag(xmltree, schema_dict, '/fleurInput/atomSpecies/species', '/fleurInput/atomSpecies/species',
                       'energyParameters', changes)

    nodes = eval_xpath(root, '/fleurInput/atomSpecies/species/energyParameters')

    assert [node.attrib.items() for node in nodes] == [[(key, str(val)) for key, val in changes.items()]] * 2


def test_xml_set_simple_tag_multiple_dict(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_xpaths import xml_set_simple_tag

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    changes = {'type': 'SCLO', 'n': 21, 'l': 3, 'eDeriv': 0}

    xml_set_simple_tag(xmltree, schema_dict, '/fleurInput/atomSpecies/species', '/fleurInput/atomSpecies/species', 'lo',
                       changes)

    nodes = eval_xpath(root, '/fleurInput/atomSpecies/species/lo')

    assert [node.attrib.items() for node in nodes] == [[(key, str(val)) for key, val in changes.items()]] * 2


def test_xml_set_simple_tag_multiple_list(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_xpaths import xml_set_simple_tag

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    changes = [{'type': 'SCLO', 'n': 21, 'l': 3, 'eDeriv': 0}, {'type': 'SCLO', 'n': 22, 'l': 4, 'eDeriv': 0}]

    xml_set_simple_tag(xmltree, schema_dict, '/fleurInput/atomSpecies/species', '/fleurInput/atomSpecies/species', 'lo',
                       changes)

    nodes = eval_xpath(root, '/fleurInput/atomSpecies/species/lo')

    assert [node.attrib.items() for node in nodes
            ] == [[(key, str(val)) for key, val in change.items()] for change in changes] * 2


def test_xml_set_simple_tag_differing_xpaths(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_xpaths import xml_set_simple_tag

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    changes = [{'type': 'SCLO', 'n': 21, 'l': 3, 'eDeriv': 0}, {'type': 'SCLO', 'n': 22, 'l': 4, 'eDeriv': 0}]

    xml_set_simple_tag(xmltree, schema_dict, "/fleurInput/atomSpecies/species[@name='Fe-1']",
                       '/fleurInput/atomSpecies/species', 'lo', changes)

    nodes = eval_xpath(root, '/fleurInput/atomSpecies/species/lo')

    assert [node.attrib.items() for node in nodes] == [[('type', 'SCLO'), ('n', '21'), ('l', '3'), ('eDeriv', '0')],
                                                       [('type', 'SCLO'), ('n', '22'), ('l', '4'), ('eDeriv', '0')],
                                                       [('type', 'SCLO'), ('l', '1'), ('n', '5'), ('eDeriv', '0')]]


def test_xml_set_simple_tag_create_single(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_xpaths import xml_set_simple_tag

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    changes = {'ne': 6}

    with pytest.raises(ValueError, match="Could not create tag 'realAxis' because atleast one subtag is missing"):
        xml_set_simple_tag(xmltree, schema_dict, '/fleurInput/calculationSetup/greensFunction',
                           '/fleurInput/calculationSetup/greensFunction', 'realAxis', changes)

    xml_set_simple_tag(xmltree,
                       schema_dict,
                       '/fleurInput/calculationSetup/greensFunction',
                       '/fleurInput/calculationSetup/greensFunction',
                       'realAxis',
                       changes,
                       create_parents=True)
    node = eval_xpath(root, '/fleurInput/calculationSetup/greensFunction/realAxis')

    assert node.attrib.items() == [('ne', '6')]


def test_xml_set_simple_tag_create_multiple(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_xpaths import xml_set_simple_tag

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    changes = [{'n': 6}, {'n': 7}]

    with pytest.raises(ValueError,
                       match="Could not create tag 'contourSemicircle' because atleast one subtag is missing"):
        xml_set_simple_tag(xmltree, schema_dict, '/fleurInput/calculationSetup/greensFunction',
                           '/fleurInput/calculationSetup/greensFunction', 'contourSemicircle', changes)

    xml_set_simple_tag(xmltree,
                       schema_dict,
                       '/fleurInput/calculationSetup/greensFunction',
                       '/fleurInput/calculationSetup/greensFunction',
                       'contourSemicircle',
                       changes,
                       create_parents=True)
    nodes = eval_xpath(root, '/fleurInput/calculationSetup/greensFunction/contourSemicircle')

    assert [node.attrib.items() for node in nodes] == [[('n', '6')], [('n', '7')]]


def test_xml_set_complex_tag(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_xpaths import xml_set_complex_tag

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    changes = {'jspins': 4, 'mtNocoParams': {'l_mtNocoPot': True}, 'qss': [0.0, 1.0, 2.0]}

    xml_set_complex_tag(xmltree, schema_dict, '/fleurInput/calculationSetup/magnetism',
                        '/fleurInput/calculationSetup/magnetism', changes)

    assert str(eval_xpath(root, '/fleurInput/calculationSetup/magnetism/@jspins')) == '4'
    assert str(eval_xpath(root, '/fleurInput/calculationSetup/magnetism/mtNocoParams/@l_mtNocoPot')) == 'T'
    assert str(eval_xpath(
        root,
        '/fleurInput/calculationSetup/magnetism/qss/text()')) == ' 0.0000000000000  1.0000000000000  2.0000000000000'


def test_xml_set_complex_tag_differing_xpaths_recursive_complex_single(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_xpaths import xml_set_complex_tag

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    changes = {'electronConfig': {'valenceConfig': 'TEST'}}

    xml_set_complex_tag(xmltree, schema_dict, "/fleurInput/atomSpecies/species[@name='Fe-1']",
                        '/fleurInput/atomSpecies/species', changes)

    nodes = eval_xpath(root, '/fleurInput/atomSpecies/species/electronConfig/valenceConfig/text()')
    assert [str(node) for node in nodes] == ['TEST', '(5p1/2) (5p3/2) (6s1/2) (5d3/2) (5d5/2)']


def test_xml_set_complex_tag_recursive_complex_multiple(load_inpxml):

    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.xml_setters_xpaths import xml_set_complex_tag

    xmltree, schema_dict = load_inpxml(TEST_INPXML_PATH)
    root = xmltree.getroot()

    changes = {'ldaHIA': [{'addArg': {'key': 'TEST'}}, {'addArg': {'key': 'TEST2'}}]}

    xml_set_complex_tag(xmltree, schema_dict, "/fleurInput/atomSpecies/species[@name='Fe-1']",
                        '/fleurInput/atomSpecies/species', changes)

    nodes = eval_xpath(root, '/fleurInput/atomSpecies/species/ldaHIA/addArg/@key')
    assert [str(node) for node in nodes] == ['TEST', 'TEST2']



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the masci-tools package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/JuDFTteam/masci-tools    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit                                        #
###############################################################################
'''
masci-tools
'''
import logging

__version__ = '0.4.3'

logging.getLogger(__name__).addHandler(logging.NullHandler())



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
This module defines subclasses of UserDict and UserList to be able to prevent
unintended modifications
"""
from collections import UserDict, UserList
from contextlib import contextmanager


@contextmanager
def LockContainer(lock_object):
    """
    Contextmanager for temporarily locking a lockable object. Object is unfrozen
    when exiting with block

    :param lock_object: lockable container (not yet frzen)
    """

    assert isinstance(lock_object, (LockableDict, LockableList)), f'Wrong type Got: {lock_object.__class__}'

    assert not lock_object._locked, f'{lock_object.__class__.__name__} was already locked before entering the contextmanager'

    lock_object.freeze()

    try:
        yield
    finally:
        if isinstance(lock_object, LockableDict):
            lock_object._LockableDict__unfreeze()
        elif isinstance(lock_object, LockableList):
            lock_object._LockableList__unfreeze()


class LockableDict(UserDict):
    """
    Subclass of UserDict, which can prevent modifications to itself.
    Raises `RuntimeError` if modification is attempted.

    Use :py:meth:`LockableDict.freeze()` to enforce. :py:meth:`LockableDict.get_unlocked()`
    returns a copy of the locked object with builtin lists and dicts

    :param recursive: bool if True (default) all subitems (lists or dicts) are converted into their
                      lockable counterparts

    All other args or kwargs will be passed on to initialize the `UserDict`

    IMPORTANT NOTE:
        This is not a direct subclass of dict. So isinstance(a, dict)
        will be False if a is an LockableDict

    """

    def __init__(self, *args, recursive=True, **kwargs):
        self._locked = False
        self._recursive = recursive
        super().__init__(*args, **kwargs)

    def __check_lock(self):
        if self._locked:
            raise RuntimeError('Modification not allowed')

    def __delitem__(self, key):
        self.__check_lock()
        super().__delitem__(key)

    def __setitem__(self, key, value):
        self.__check_lock()
        if isinstance(value, list):
            super().__setitem__(key, LockableList(value, recursive=self._recursive))
        elif isinstance(value, dict):
            super().__setitem__(key, LockableDict(value, recursive=self._recursive))
        else:
            super().__setitem__(key, value)

    def freeze(self):
        """
        Freezes the object. This prevents further modifications
        """
        self.__freeze()

    def __freeze(self):

        if self._recursive:
            for key, val in self.items():
                if isinstance(val, LockableDict):
                    val.__freeze()
                elif isinstance(val, LockableList):
                    val._LockableList__freeze()

        self._locked = True

    def __unfreeze(self):

        if self._recursive:
            for key, val in self.items():
                if isinstance(val, LockableDict):
                    val.__unfreeze()
                elif isinstance(val, LockableList):
                    val._LockableList__unfreeze()

        self._locked = False

    def get_unlocked(self):
        """
        Get copy of object with builtin lists and dicts
        """
        if self._recursive:
            ret_dict = {}
            for key, value in self.items():
                if isinstance(value, LockableDict):
                    ret_dict[key] = value.get_unlocked()
                elif isinstance(value, LockableList):
                    ret_dict[key] = value.get_unlocked()
                else:
                    ret_dict[key] = value
        else:
            ret_dict = dict(self)

        return ret_dict


class LockableList(UserList):
    """
    Subclass of UserList, which can prevent modifications to itself.
    Raises `RuntimeError` if modification is attempted.

    Use :py:meth:`LockableList.freeze()` to enforce. :py:meth:`LockableList.get_unlocked()`
    returns a copy of the locked object with builtin lists and dicts

    :param recursive: bool if True (default) all subitems (lists or dicts) are converted into their
                      lockable counterparts

    All other args or kwargs will be passed on to initialize the `UserList`

    IMPORTANT NOTE:
        This is not a direct subclass of list. So isinstance(a, list)
        will be False if a is an LockableList

    """

    def __init__(self, *args, recursive=True, **kwargs):
        self._locked = False
        self._recursive = recursive
        super().__init__(*args, **kwargs)
        if self._recursive:
            #Convert sublists and subdicts into Lockable counterparts (super__init__ just copies the values)
            for indx, item in enumerate(self):
                if isinstance(item, list):
                    super().__setitem__(indx, LockableList(item, recursive=self._recursive))
                elif isinstance(item, dict):
                    super().__setitem__(indx, LockableDict(item, recursive=self._recursive))

    def __check_lock(self):
        if self._locked:
            raise RuntimeError('Modification not allowed')

    def __delitem__(self, i):
        self.__check_lock()
        super().__delitem__(i)

    def __setitem__(self, i, item):
        self.__check_lock()
        if isinstance(item, list):
            super().__setitem__(i, LockableList(item, recursive=self._recursive))
        elif isinstance(item, dict):
            super().__setitem__(i, LockableDict(item, recursive=self._recursive))
        else:
            super().__setitem__(i, item)

    def __iadd__(self, other):
        self.__check_lock()
        return super().__iadd__(other)

    def __imul__(self, n):
        self.__check_lock()
        return super().__imul__(n)

    def append(self, item):
        self.__check_lock()
        super().append(item)

    def insert(self, i, item):
        self.__check_lock()
        super().insert(i, item)

    def pop(self, i=-1):
        """
        return the value at index i (default last) and remove it from list
        """
        self.__check_lock()
        super().pop(i=i)

    def remove(self, item):
        self.__check_lock()
        super().remove(item)

    def clear(self):
        """
        Clear the list
        """
        self.__check_lock()
        super().clear()

    def reverse(self):
        self.__check_lock()
        super().reverse()

    def sort(self, *args, **kwargs):
        self.__check_lock()
        super().sort(*args, **kwargs)

    def extend(self, other):
        self.__check_lock()
        super().extend(other)

    def freeze(self):
        """
        Freezes the object. This prevents further modifications
        """
        self.__freeze()

    def __freeze(self):

        if self._recursive:
            for val in self:
                if isinstance(val, LockableList):
                    val.__freeze()
                elif isinstance(val, LockableDict):
                    val._LockableDict__freeze()

        self._locked = True

    def __unfreeze(self):

        if self._recursive:
            for val in self:
                if isinstance(val, LockableList):
                    val.__unfreeze()
                elif isinstance(val, LockableDict):
                    val._LockableDict__unfreeze()
        self._locked = False

    def get_unlocked(self):
        """
        Get copy of object with builtin lists and dicts
        """
        if self._recursive:
            ret_list = []
            for value in self:
                if isinstance(value, LockableDict):
                    ret_list.append(value.get_unlocked())
                elif isinstance(value, LockableList):
                    ret_list.append(value.get_unlocked())
                else:
                    ret_list.append(value)
        else:
            ret_list = list(self)

        return ret_list



# -*- coding: utf-8 -*-
"""
Tests of the plotter base class
"""
from masci_tools.vis import Plotter, ensure_plotter_consistency
import pytest

TEST_DICT = {'A': {'test1': 12, 'test2': 4}, 'B': 3.0, 'C': 'title'}


def test_plotter_access():
    """
   Test basic item acces and make sure the right structure was created
   """
    p = Plotter(TEST_DICT)

    assert dict(p._params) == TEST_DICT

    #Test basic getitem method
    for key, value in TEST_DICT.items():
        assert p[key] == value


def test_plotter_set_parameters():
    """
   Test setting of parameters
   """
    p = Plotter(TEST_DICT)

    p['B'] = 0

    assert p['B'] == 0

    with pytest.raises(KeyError):
        p['D'] = 'not a key'

    p['A'] = {'test3': 'extra'}

    assert p['A'] == {'test1': 12, 'test2': 4, 'test3': 'extra'}

    with pytest.raises(ValueError):
        p['A'] = 2.0

    assert dict(p._params.parents) == TEST_DICT
    p.reset_parameters()

    assert dict(p._params) == TEST_DICT


def test_plotter_set_parameters_function():
    """
   Test setting parameters via the provided function
   """
    p = Plotter(TEST_DICT)

    p.set_parameters(continue_on_error=True, **{'B': 0, 'A': {'test3': 'extra'}, 'D': 'not a key'})

    assert dict(p._params) == {'A': {'test1': 12, 'test2': 4, 'test3': 'extra'}, 'B': 0, 'C': 'title'}

    p.reset_parameters()

    with pytest.raises(KeyError):
        p.set_parameters(**{'B': 0, 'A': {'test3': 'extra'}, 'D': 'not a key'})

    assert dict(p._params) == TEST_DICT


def test_plotter_set_defaults():
    """
   Test setting of defaults
   """

    p = Plotter(TEST_DICT)

    p.set_defaults(continue_on_error=True, **{'B': 0, 'A': {'test3': 'extra'}, 'D': 'not a key'})

    expected_result = {'A': {'test1': 12, 'test2': 4, 'test3': 'extra'}, 'B': 0, 'C': 'title'}

    assert dict(p._params) == expected_result
    assert dict(p._params.parents) == expected_result

    p.reset_defaults()

    assert dict(p._params) == TEST_DICT

    with pytest.raises(KeyError):
        p.set_defaults(**{'B': 0, 'A': {'test3': 'extra'}, 'D': 'not a key'})

    assert dict(p._params) == TEST_DICT


def test_plotter_init_defaults():
    """
   Test setting defaults via initialization
   """
    p = Plotter(TEST_DICT, **{'B': 0, 'A': {'test3': 'extra'}, 'D': 'not a key'})

    expected_result = {'A': {'test1': 12, 'test2': 4, 'test3': 'extra'}, 'B': 0, 'C': 'title'}

    assert dict(p._params) == expected_result
    assert dict(p._params.parents) == expected_result


def test_plotter_add_parameter():
    """
    Test adding of custom parameters
    """

    p = Plotter(TEST_DICT)

    p.add_parameter('D')
    assert p['D'] is None
    p['D'] = 2
    assert p['D'] == 2

    p.add_parameter('E', default_from='C')
    assert p['E'] == 'title'

    p.remove_added_parameters()
    p.reset_parameters()

    assert dict(p._params) == TEST_DICT


def test_plotter_decorator_working():
    """
    Test the ensure_plotter_consistency decorator for a working function
    """

    p = Plotter(TEST_DICT)

    @ensure_plotter_consistency(p)
    def test_function():
        p.set_parameters(B=4.0)

    test_function()

    assert dict(p._params) == TEST_DICT


def test_plotter_decorator_raised_error():
    """
    Test the ensure_plotter_consistency decorator for a function with an error occuring durin execution
    """

    p = Plotter(TEST_DICT)

    @ensure_plotter_consistency(p)
    def test_function():
        p.set_parameters(B=4.0)
        raise ValueError('Test')

    with pytest.raises(ValueError, match='Test'):
        test_function()

    assert dict(p._params) == TEST_DICT


def test_plotter_decorator_set_defaults():
    """
    Test the ensure_plotter_consistency decorator for a function changing the defaults (raises Error)
    """

    p = Plotter(TEST_DICT)

    @ensure_plotter_consistency(p)
    def test_function():
        p.set_defaults(C='TEST')
        p.set_parameters(B=4.0)

    with pytest.raises(ValueError):
        test_function()

    assert dict(p._params) == TEST_DICT


def test_plotter_decorator_add_parameter():
    """
    Test the ensure_plotter_consistency decorator for a function adding custom parameters
    """

    p = Plotter(TEST_DICT)

    @ensure_plotter_consistency(p)
    def test_function():
        p.add_parameter('D')
        p.add_parameter('E', default_from='C')

    test_function()

    assert dict(p._params) == TEST_DICT


def test_plotter_decorator_add_parameter_raised_error():
    """
    Test the ensure_plotter_consistency decorator for a function adding custom parameters with an exception occuring
    """

    p = Plotter(TEST_DICT)

    @ensure_plotter_consistency(p)
    def test_function():
        p.add_parameter('D')
        p.add_parameter('E', default_from='C')
        raise ValueError('Test')

    with pytest.raises(ValueError, match='Test'):
        test_function()

    assert dict(p._params) == TEST_DICT


WORKING_VALUES = [[None, 3, None], 'Test', [1, 2, 3, 4, 5], {4: 'Test2'}, [5.0], {'NotAList': 'Test2'}]
GIVEN_NUM_PLOTS = [5, 3, 5, 5, 2, 1]
SINGLE_PLOT_ERROR = [True, False, True, True, True, False]
EXPECTED_RESULTS = [[None, 3, None, None, None], 'Test', [1, 2, 3, 4, 5], [None, None, None, None, 'Test2'],
                    [5.0, None], {
                        'NotAList': 'Test2'
                    }]
EXPECTED_RESULTS_WITH_DEFAULT = [['default', 3, 'default', 'default', 'default'], 'Test', [1, 2, 3, 4, 5],
                                 ['default', 'default', 'default', 'default', 'Test2'], [5.0, 'default'], {
                                     'NotAList': 'Test2'
                                 }]


@pytest.mark.parametrize('given_value,num_plots,expected_list', zip(WORKING_VALUES, GIVEN_NUM_PLOTS, EXPECTED_RESULTS))
def test_plotter_convert_to_complete_list_multiple_plots_list(given_value, num_plots, expected_list):

    complete_list = Plotter.convert_to_complete_list(given_value, False, num_plots)

    assert complete_list == expected_list


@pytest.mark.parametrize('given_value,num_plots,expected_list',
                         zip(WORKING_VALUES, GIVEN_NUM_PLOTS, EXPECTED_RESULTS_WITH_DEFAULT))
def test_plotter_convert_to_complete_list_multiple_plots_list_default(given_value, num_plots, expected_list):

    complete_list = Plotter.convert_to_complete_list(given_value, False, num_plots, default='default')

    assert complete_list == expected_list


@pytest.mark.parametrize('given_value, error_expected, expected_result',
                         zip(WORKING_VALUES, SINGLE_PLOT_ERROR, EXPECTED_RESULTS))
def test_plotter_convert_to_complete_list_single_plot_no_list_allowed(given_value, error_expected, expected_result):

    if error_expected:
        with pytest.raises(ValueError):
            res = Plotter.convert_to_complete_list(given_value, True, 1)
    else:
        res = Plotter.convert_to_complete_list(given_value, True, 1)

        assert res == expected_result



# -*- coding: utf-8 -*-
"""
Tests for fleur outxml_parser specific conversion functions
"""
import pytest
import logging

LOGGER = logging.getLogger(__name__)

TEST_DICTS = [{
    'end_date': {
        'date': '2020/12/10',
        'time': '16:51:35'
    },
    'start_date': {
        'date': '2020/12/10',
        'time': '16:51:21'
    }
}, {
    'end_date': {
        'date': '2020/12/10',
        'time': '16:58:49'
    },
    'start_date': {
        'date': '2020/12/10',
        'time': '16:51:21'
    }
}, {
    'end_date': {
        'date': '2020/12/10',
        'time': '22:31:67'
    },
    'start_date': {
        'date': '2020/12/10',
        'time': '16:51:21'
    }
}, {
    'end_date': {
        'date': '2020/12/11',
        'time': '16:51:35'
    },
    'start_date': {
        'date': '2020/12/10',
        'time': '16:51:35'
    }
}, {
    'end_date': {
        'date': '2020/12/31',
        'time': '09:12:45'
    },
    'start_date': {
        'date': '2020/12/10',
        'time': '16:51:21'
    }
}, {
    'end_date': {
        'date': None,
        'time': None
    },
    'start_date': {
        'date': '2020/12/10',
        'time': '16:51:21'
    }
}, {
    'end_date': {
        'date': '2020/12/10',
        'time': '16:51:35'
    },
    'start_date': {
        'date': None,
        'time': None
    }
}, {
    'end_date': {
        'date': None,
        'time': None
    },
    'start_date': {
        'date': None,
        'time': None
    }
}]

TEST_WALLTIMES = [14, 448, 20446, 86400, 1786884, -60681, 60695, 0]

TEST_WARNINGS = [[], [], [], [], [],
                 [
                     'Endtime was unparsed, inp.xml prob not complete, do not believe the walltime!',
                     'Enddate was unparsed, inp.xml prob not complete, do not believe the walltime!'
                 ],
                 [
                     'Starttime was unparsed, inp.xml prob not complete, do not believe the walltime!',
                     'Startdate was unparsed, inp.xml prob not complete, do not believe the walltime!'
                 ],
                 [
                     'Starttime was unparsed, inp.xml prob not complete, do not believe the walltime!',
                     'Endtime was unparsed, inp.xml prob not complete, do not believe the walltime!',
                     'Startdate was unparsed, inp.xml prob not complete, do not believe the walltime!',
                     'Enddate was unparsed, inp.xml prob not complete, do not believe the walltime!'
                 ]]


@pytest.mark.parametrize('input_dict, walltime, warnings', zip(TEST_DICTS, TEST_WALLTIMES, TEST_WARNINGS))
def test_calculate_walltime(caplog, input_dict, walltime, warnings):
    """
   Test of the calculate_walltime function
   """
    from masci_tools.io.parsers.fleur.outxml_conversions import calculate_walltime

    with caplog.at_level(logging.WARNING):
        out_dict = calculate_walltime(input_dict, logger=LOGGER)

        assert out_dict['walltime_units'] == 'seconds'
        assert out_dict['walltime'] == walltime

    if len(warnings) == 0:
        assert caplog.text == ''
    else:
        for expected_warning in warnings:
            assert expected_warning in caplog.text



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
In this module migration functions for the task definitions are collected
"""
import copy
from masci_tools.util.parse_tasks_decorators import register_migration


@register_migration(base_version='0.33', target_version=['0.31', '0.30', '0.29'])
def migrate_033_to_031(definition_dict):
    """
    Migrate definitions for MaX5 release to MaX4 release

    Changes:
        - LDA+U density matrix distance output did not exist
    """

    new_dict = copy.deepcopy(definition_dict)
    new_dict.pop('nmmp_distances')

    return new_dict


@register_migration(base_version='0.34', target_version='0.33')
def migrate_034_to_033(definition_dict):
    """
    Migrate definitions for MaX5 bugfix release to MaX5 release

    Changes:
        - forcetheorem units attribute did not exist (get from 'sumValenceSingleParticleEnergies')
    """

    new_dict = copy.deepcopy(definition_dict)

    force_units = {
        'parse_type': 'attrib',
        'path_spec': {
            'name': 'units',
            'tag_name': 'sumValenceSingleParticleEnergies'
        }
    }

    new_dict['forcetheorem_mae']['mae_force_units'] = copy.deepcopy(force_units)
    new_dict['forcetheorem_ssdisp']['spst_force_units'] = copy.deepcopy(force_units)
    new_dict['forcetheorem_jij']['jij_force_units'] = copy.deepcopy(force_units)
    new_dict['forcetheorem_dmi']['dmi_force_units'] = copy.deepcopy(force_units)

    return new_dict



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), 2018 Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.    #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools    #
# For further information on the license, see the LICENSE.txt file            #
#                                                                             #
###############################################################################

__version__ = 0.1


def read_shapefun(path='.'):
    """
  Read vertices of shapefunctions with
  Zoom into shapefun of a single atom

  :author: Philipp Ruessmann

  :param path: path where voronoi output is found (optional, defaults to './')

  :returns pos: positions of the centers of the shapefunctions
  :returns out: dictionary of the vertices of the shapefunctions
  """
    from masci_tools.io.common_functions import search_string
    from numpy import array

    path += '/'

    # first read vertices file
    with open(path + 'vertices.dat') as f:
        out = {}
        face, face_old = -1, -1
        for line in f:
            if '# serial:' in line[:9]:
                # skip line with serial number
                pass
            else:
                if 'representative' in line:
                    atm = int(line.split()[-1])
                    i_newat = 1
                elif 'Face' in line:
                    i_newat = 0
                    face = int(line.split()[-1])
                    i_newface = 1
                    tmp = []
                else:
                    i_newface = 0
                    tmp2 = line.split()
                    if tmp2 != []:
                        tmp2 = [float(i) for i in tmp2]
                        tmp.append(tmp2)
                    atm_old, face_old = atm, face
                if face_old != face:
                    out[atm, face] = tmp

    # then read positions from inputcard
    with open(path + 'inputcard') as file:
        inp = file.readlines()
        # convert to uppercase
        for index, line in enumerate(inp):
            inp[index] = line.upper()
        # search number of atoms in inputcard
        iline = search_string('NAEZ', inp)
        if iline < 0:
            iline = search_string('NATYP', inp)
            if iline == -1:
                raise ValueError('Error reading NAEZ or NATYP from inputcard.')
        naez = int(inp.pop(iline).split('=')[1].split()[0])

        # read rbasis positions
        iline = search_string('RBASIS', inp)
        if iline >= 0:
            pos = []
            for iatom in range(naez):
                pos.append(inp.pop(iline + 1))
            # finally convert to float arrays
            pos = array([[float(j) for j in i.split()[:3]] for i in pos])
        else:
            raise ValueError('Error reading RBASIS from inputcard.')

    return pos, out



# -*- coding: utf-8 -*-
"""
Tests of the bokeh visualization. Since the concrete visualization is difficult
to test we check the content of the underlying json for correctness
"""
from bokeh.io import curdoc
import numpy as np
import pandas as pd


def prepare_for_regression(data):
    """
    Make the dict form the produced json data
    suitable for data_regression

    - remove any reference to ids
    - sort lists after types and given attributes for reproducible order

    :param data: dict with the json data produced for the bokeh figure
    """

    for key, val in list(data.items()):
        if key in ('id', 'root_ids'):
            del data[key]
        elif isinstance(val, dict):
            data[key] = prepare_for_regression(val)
        elif isinstance(val, list):
            for index, entry in enumerate(val):
                if isinstance(entry, dict):
                    val[index] = prepare_for_regression(entry)
            if all(isinstance(x, dict) for x in val):
                data[key] = sorted(val, key=lambda x: (x['type'], *x.get('attributes', {}).items()))
            else:
                data[key] = val

    return data


class TestBokehScatter:  #pylint: disable=missing-class-docstring

    def test_default(self, data_regression):
        """
        Test with default values
        """
        from masci_tools.vis.bokeh_plots import bokeh_scatter
        x = np.linspace(-10, 10, 100)
        y = x**2

        source = pd.DataFrame(data={'x': x, 'y': y})

        p = bokeh_scatter(source, show=False)

        curdoc().clear()
        curdoc().add_root(p)
        data_regression.check(prepare_for_regression(curdoc().to_json()))



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), 2018 Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.    #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools    #
# For further information on the license, see the LICENSE.txt file            #
#                                                                             #
###############################################################################

from __future__ import print_function
from __future__ import absolute_import
from six.moves import range


def plot_shapefun(pos, out, mode):
    """
  Creates a simple matplotlib image to show the shapefunctions given it's positions in the unit cell, the atoms's vertices in `ut` and the plotting mode

  :author: Philipp Ruessmann

  :param pos: positions of the centers of the cells
  :param verts: array of vertices of the shapefunction (outlines of shapes)
  :param mode: 'all' or 'single' determines whether or not all shapes are combined in a single figure or plotted as individual figures

  :returns ax: return the axis in which the plot was done (useful to pass to 'change_zoom' and 'zoom_in' functions of this module
  """
    import matplotlib.pyplot as plt
    from mpl_toolkits.mplot3d import Axes3D
    from numpy import array, shape

    if mode == 'all':
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')

    Natm = len(pos)

    for j in range(1, Natm + 1):
        if mode == 'single':
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')
        for i in list(out.keys()):
            if i[0] == j:
                linecolor = 'r'
            if j in [2, 3, 6, 7]:
                linecolor = 'b'
            d = array(out[i])
            if len(shape(pos)) > 1 and len(pos) > 2 and mode == 'all':
                dx, dy, dz = pos[j - 1, 0], pos[j - 1, 1], pos[j - 1, 2]
            else:
                dx, dy, dz = 0, 0, 0
            ax.plot(d[:, 0] + dx, d[:, 1] + dy, d[:, 2] + dz, linecolor)

    if mode == 'all':
        xy_scale = 1
        scale = 1
        ax.set_xlim(-scale * xy_scale, scale * xy_scale)
        ax.set_ylim(-scale * xy_scale, scale * xy_scale)
        ax.set_zlim(-scale, scale)

    return ax


def change_zoom(ax, zoom_range, center=[0, 0, 0]):
    """
  Change the zoom of a 3d plot

  :author: Philipp Ruessmann

  :param ax: axis which is zoomed
  :param zoom_range: range to which the image is zoomed, total range from center-zoom_range to center+zoom_range
  :param center: center of the zoomed region (optional, defaults to origin)
  """
    ax.set_xlim(center[0] - zoom_range, center[0] + zoom_range)
    ax.set_ylim(center[1] - zoom_range, center[1] + zoom_range)
    ax.set_zlim(center[2] - zoom_range, center[2] + zoom_range)


def zoom_in(ax, atm, pos, zoom_range=10):
    """
  Zoom into shapefun of a single atom

  :author: Philipp Ruessmann

  :param ax: axis in which shapefun plot is found
  :param atm: atom index whose shapefunction is zoomed
  :param pos: array of positions of centers of the shapes (needed to shift center of zommed region to correct atom
  :param zoom_range: range of the zoomed region (optional, defaults to 10)
  """
    center = pos[atm - 1]
    print('zoom in to atom', atm, ':', center)
    change_zoom(ax, zoom_range, center)



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
This module provides the functionality to create/load the schema_dict for the
FleurInputSchema.xsd
"""
from .fleur_schema_parser_functions import *  #pylint: disable=unused-wildcard-import
from masci_tools.util.xml.common_functions import clear_xml
from masci_tools.util.case_insensitive_dict import CaseInsensitiveDict
from lxml import etree
import copy
from collections import UserList


def create_outschema_dict(path, inp_path=None, inpschema_dict=None):
    """
    Creates dictionary with information about the FleurOutputSchema.xsd.
    The functions, whose results are added to the schema_dict and the corresponding keys
    are defined in schema_actions

    :param path: str path to the folder containing the FleurOutputSchema.xsd file
    :param inp_path: str path to the FleurInputSchema.xsd file (defaults to the same folder as path)
    """

    #Add new functionality to this dictionary here
    schema_actions = {
        'input_tag': get_input_tag,
        'root_tag': get_root_tag,
        '_basic_types': get_basic_types,
        'attrib_types': extract_attribute_types,
        'simple_elements': get_basic_elements,
        'tag_paths': get_tag_paths,
        'iteration_tag_paths': get_tag_paths,
        'unique_attribs': get_unique_attribs,
        'unique_path_attribs': get_unique_path_attribs,
        'other_attribs': get_other_attribs,
        'iteration_unique_attribs': get_unique_attribs,
        'iteration_unique_path_attribs': get_unique_path_attribs,
        'iteration_other_attribs': get_other_attribs,
        'tag_info': get_tag_info,
        'iteration_tag_info': get_tag_info,
        'omitt_contained_tags': get_omittable_tags,
    }

    #print(f'processing: {path}/FleurOutputSchema.xsd')
    xmlschema = etree.parse(path)
    xmlschema = clear_xml(xmlschema)

    namespaces = {'xsd': 'http://www.w3.org/2001/XMLSchema'}
    out_version = str(xmlschema.xpath('/xsd:schema/@version', namespaces=namespaces)[0])

    if inpschema_dict is not None:
        input_basic_types = inpschema_dict.get('_basic_types').get_unlocked()
    else:
        if inp_path is None:
            inp_path = path.replace('FleurOutputSchema', 'FleurInputSchema')
        #Parse type definitions directly from inputSchema
        inpxmlschema = etree.parse(inp_path)
        inpxmlschema = clear_xml(inpxmlschema)
        input_basic_types = get_basic_types(inpxmlschema, namespaces)

    schema_dict = {}
    schema_dict['out_version'] = out_version
    for key, action in schema_actions.items():
        addargs = {'input_basic_types': input_basic_types}
        if key in ['unique_attribs', 'unique_path_attribs', 'other_attribs', 'tag_paths', 'tag_info']:
            addargs['stop_iteration'] = True
        elif key in [
                'iteration_unique_attribs', 'iteration_unique_path_attribs', 'iteration_other_attribs',
                'iteration_tag_paths', 'iteration_tag_info'
        ]:
            addargs['iteration_root'] = True
            addargs['iteration'] = True
        schema_dict[key] = action(xmlschema, namespaces, **schema_dict, **addargs)

    schema_dict['_input_basic_types'] = input_basic_types

    #We cannot do the conversion to CaseInsensitiveDict before since we need the correct case
    #For these attributes in the attrib_path functions
    schema_dict['simple_elements'] = CaseInsensitiveDict(schema_dict['simple_elements'])

    return schema_dict


def merge_schema_dicts(inputschema_dict, outputschema_dict):
    """
    Merge the information from the input schema into the outputschema
    This combines the type information and adjusts the paths from the inputschema
    to be valid in the out.xml file, i.e. `/fleurInput/cell` becomes `/fleurOutput/fleurInput/cell`
    or `/fleurOutput/inputData/cell` (depending on the version of the output)

    `_basic_types` and `_input_basic_types` stay untouched since they should not be used outside
    of the construction and are merely a informational entry

    :param inputschema_dict: schema dict for the input schema
    :param outputschema_dict: schema dict for the output schema

    :returns: schema dictionary with the information merged
    """
    merged_outschema_dict = copy.deepcopy(outputschema_dict)
    merged_outschema_dict['inp_version'] = inputschema_dict['inp_version']

    input_tag_path = outputschema_dict['tag_paths'][outputschema_dict['input_tag']]
    input_root_tag = inputschema_dict['root_tag']

    #
    # 1. Merge path entries and modify the input paths
    #
    path_entries = {'tag_paths', 'unique_attribs', 'unique_path_attribs', 'other_attribs'}
    for entry in path_entries:
        for key, val in inputschema_dict[entry].items():

            paths = merged_outschema_dict[entry].get(key, UserList())

            if not isinstance(paths, UserList):
                paths = [paths]
            paths = set(paths)

            if not isinstance(val, UserList):
                val = [val]
            new_paths = {f"{input_tag_path}{inp_path.replace(f'/{input_root_tag}','')}" for inp_path in val}

            new_paths = sorted(paths.union(new_paths))
            if len(new_paths) == 1:
                merged_outschema_dict[entry][key] = new_paths[0]
            else:
                merged_outschema_dict[entry][key] = new_paths

    #Remove the root_tag of the input if it is different from the input tag in the out.xml
    if input_root_tag != outputschema_dict['input_tag']:
        merged_outschema_dict['tag_paths'].pop(input_root_tag)

    #
    # 2. Insert tag_info paths (There is no possibility for overlap here)
    #
    new_tag_info_entries = {
        f"{input_tag_path}{path.replace(f'/{input_root_tag}','')}": info
        for path, info in inputschema_dict['tag_info'].items()
    }
    merged_outschema_dict['tag_info'].update(new_tag_info_entries)

    #
    # 3. Merge the attribute type information
    #
    for attrib, types in inputschema_dict['attrib_types'].items():
        out_types = set(merged_outschema_dict['attrib_types'].get(attrib, []))

        new_types = sorted(out_types.union(set(types)))
        if 'string' in new_types:
            new_types.remove('string')
            new_types.append('string')

        merged_outschema_dict['attrib_types'][attrib] = new_types

    #
    # 4. Merge the definition information for text tags
    #
    for name, definition in inputschema_dict['simple_elements'].items():
        if name in merged_outschema_dict['simple_elements']:
            new_types = merged_outschema_dict['simple_elements'].get(name)
            for new_definition in definition:
                for index, old_dict in enumerate(new_types):
                    equal_dicts = True
                    for key, value in old_dict.items():
                        if new_definition[key] != value:
                            equal_dicts = False
                    if equal_dicts:
                        break
                    if index == len(new_types) - 1:
                        new_types.append(new_definition)
            merged_outschema_dict['simple_elements'][name] = new_types
        else:
            merged_outschema_dict['simple_elements'][name] = definition

    #
    # 5. Merge the omittable tags
    #
    merged_outschema_dict['omitt_contained_tags'] = sorted(
        set(merged_outschema_dict.get('omitt_contained_tags')).union(inputschema_dict.get('omitt_contained_tags')))

    return merged_outschema_dict



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
Functions for modifying the xml input file of Fleur utilizing the schema dict
and as little knowledge of the concrete xpaths as possible
"""
from masci_tools.util.schema_dict_util import get_tag_xpath
from masci_tools.util.schema_dict_util import get_attrib_xpath
from masci_tools.io.parsers.fleur.fleur_schema import schema_dict_version_dispatch


def create_tag(xmltree, schema_dict, tag_name, complex_xpath=None, create_parents=False, occurrences=None, **kwargs):
    """
    This method creates a tag with a uniquely identified xpath under the nodes of its parent.
    If there are no nodes evaluated the subtags can be created with `create_parents=True`

    The tag is always inserted in the correct place if a order is enforced by the schema

    :param xmltree: an xmltree that represents inp.xml
    :param schema_dict: InputSchemaDict containing all information about the structure of the input
    :param tag_name: str of the tag to create
    :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
    :param create_parents: bool optional (default False), if True and the given xpath has no results the
                           the parent tags are created recursively
    :param occurrences: int or list of int. Which occurence of the parent nodes to create a tag.
                        By default all nodes are used.

    Kwargs:
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path

    :returns: xmltree with created tags
    """
    from masci_tools.util.xml.xml_setters_xpaths import xml_create_tag_schema_dict
    from masci_tools.util.xml.common_functions import split_off_tag

    base_xpath = get_tag_xpath(schema_dict, tag_name, **kwargs)

    parent_xpath, tag_name = split_off_tag(base_xpath)

    if complex_xpath is None:
        complex_xpath = parent_xpath

    xmltree = xml_create_tag_schema_dict(xmltree,
                                         schema_dict,
                                         complex_xpath,
                                         parent_xpath,
                                         tag_name,
                                         create_parents=create_parents,
                                         occurrences=occurrences)

    return xmltree


def add_number_to_attrib(xmltree,
                         schema_dict,
                         attributename,
                         add_number,
                         complex_xpath=None,
                         mode='abs',
                         occurrences=None,
                         **kwargs):
    """
    Adds a given number to the attribute value in a xmltree specified by the name of the attribute
    and optional further specification
    If there are no nodes under the specified xpath an error is raised

    :param xmltree: an xmltree that represents inp.xml
    :param schema_dict: InputSchemaDict containing all information about the structure of the input
    :param attributename: the attribute name to change
    :param add_number: number to add/multiply with the old attribute value
    :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
    :param mode: str (either `rel` or `abs`).
                 `rel` multiplies the old value with `add_number`
                 `abs` adds the old value and `add_number`
    :param occurrences: int or list of int. Which occurence of the node to set. By default all are set.

    Kwargs:
        :param tag_name: str, name of the tag where the attribute should be parsed
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path
        :param exclude: list of str, here specific types of attributes can be excluded
                        valid values are: settable, settable_contains, other

    :returns: xmltree with shifted attribute
    """
    from masci_tools.util.xml.xml_setters_xpaths import xml_add_number_to_attrib
    from masci_tools.util.xml.common_functions import split_off_attrib

    attrib_xpath = get_attrib_xpath(schema_dict, attributename, **kwargs)

    base_xpath, attributename = split_off_attrib(attrib_xpath)

    if complex_xpath is None:
        complex_xpath = base_xpath

    xmltree = xml_add_number_to_attrib(xmltree,
                                       schema_dict,
                                       complex_xpath,
                                       base_xpath,
                                       attributename,
                                       add_number,
                                       mode=mode,
                                       occurrences=occurrences)

    return xmltree


def add_number_to_first_attrib(xmltree,
                               schema_dict,
                               attributename,
                               add_number,
                               complex_xpath=None,
                               mode='abs',
                               **kwargs):
    """
    Adds a given number to the first occurrence of an attribute value in a xmltree specified by the name of the attribute
    and optional further specification
    If there are no nodes under the specified xpath an error is raised

    :param xmltree: an xmltree that represents inp.xml
    :param schema_dict: InputSchemaDict containing all information about the structure of the input
    :param attributename: the attribute name to change
    :param add_number: number to add/multiply with the old attribute value
    :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
    :param mode: str (either `rel` or `abs`).
                 `rel` multiplies the old value with `add_number`
                 `abs` adds the old value and `add_number`

    Kwargs:
        :param tag_name: str, name of the tag where the attribute should be parsed
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path
        :param exclude: list of str, here specific types of attributes can be excluded
                        valid values are: settable, settable_contains, other

    :returns: xmltree with shifted attribute
    """
    return add_number_to_attrib(xmltree,
                                schema_dict,
                                attributename,
                                add_number,
                                complex_xpath=complex_xpath,
                                mode=mode,
                                occurrences=0,
                                **kwargs)


def set_attrib_value(xmltree,
                     schema_dict,
                     attributename,
                     attribv,
                     complex_xpath=None,
                     occurrences=None,
                     create=False,
                     **kwargs):
    """
    Sets an attribute in a xmltree to a given value, specified by its name and further
    specifications.
    If there are no nodes under the specified xpath a tag can be created with `create=True`.
    The attribute values are converted automatically according to the types of the attribute
    with :py:func:`~masci_tools.util.xml.converters.convert_attribute_to_xml()` if they
    are not `str` already.

    :param xmltree: an xmltree that represents inp.xml
    :param schema_dict: InputSchemaDict containing all information about the structure of the input
    :param attributename: the attribute name to set
    :param attribv: value or list of values to set
    :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
    :param occurrences: int or list of int. Which occurence of the node to set. By default all are set.
    :param create: bool optional (default False), if True the tag is created if is missing

    Kwargs:
        :param tag_name: str, name of the tag where the attribute should be parsed
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path
        :param exclude: list of str, here specific types of attributes can be excluded
                        valid values are: settable, settable_contains, other

    :returns: xmltree with set attribute
    """
    from masci_tools.util.xml.xml_setters_xpaths import xml_set_attrib_value
    from masci_tools.util.xml.common_functions import split_off_attrib

    #Special case for xcFunctional
    #(Also implemented here to not confuse users since it would only work in set_inpchanges otherwise)
    if attributename == 'xcFunctional':
        attributename = 'name'
        if 'exclude' not in kwargs:
            kwargs['exclude'] = ['other']
        elif 'other' not in kwargs['exclude']:
            kwargs['exclude'].append('other')

    base_xpath = get_attrib_xpath(schema_dict, attributename, **kwargs)

    base_xpath, attributename = split_off_attrib(base_xpath)

    if complex_xpath is None:
        complex_xpath = base_xpath

    xmltree = xml_set_attrib_value(xmltree,
                                   schema_dict,
                                   complex_xpath,
                                   base_xpath,
                                   attributename,
                                   attribv,
                                   occurrences=occurrences,
                                   create=create)

    return xmltree


def set_first_attrib_value(xmltree, schema_dict, attributename, attribv, complex_xpath=None, create=False, **kwargs):
    """
    Sets the first occurrence of an attribute in a xmltree to a given value, specified by its name and further
    specifications.
    If there are no nodes under the specified xpath a tag can be created with `create=True`.
    The attribute values are converted automatically according to the types of the attribute
    with :py:func:`~masci_tools.util.xml.converters.convert_attribute_to_xml()` if they
    are not `str` already.

    :param xmltree: an xmltree that represents inp.xml
    :param schema_dict: InputSchemaDict containing all information about the structure of the input
    :param attributename: the attribute name to set
    :param attribv: value or list of values to set
    :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
    :param create: bool optional (default False), if True the tag is created if is missing

    Kwargs:
        :param tag_name: str, name of the tag where the attribute should be parsed
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path
        :param exclude: list of str, here specific types of attributes can be excluded
                        valid values are: settable, settable_contains, other

    :returns: xmltree with set attribute
    """
    return set_attrib_value(xmltree,
                            schema_dict,
                            attributename,
                            attribv,
                            complex_xpath=complex_xpath,
                            create=create,
                            occurrences=0,
                            **kwargs)


def set_text(xmltree, schema_dict, tag_name, text, complex_xpath=None, occurrences=None, create=False, **kwargs):
    """
    Sets the text on tags in a xmltree to a given value, specified by the name of the tag and
    further specifications. By default the text will be set on all nodes returned for the specified xpath.
    If there are no nodes under the specified xpath a tag can be created with `create=True`.
    The text values are converted automatically according to the types
    with :py:func:`~masci_tools.util.xml.converters.convert_text_to_xml()` if they
    are not `str` already.

    :param xmltree: an xmltree that represents inp.xml
    :param schema_dict: InputSchemaDict containing all information about the structure of the input
    :param tag_name: str name of the tag, where the text should be set
    :param text: value or list of values to set
    :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
    :param occurrences: int or list of int. Which occurence of the node to set. By default all are set.
    :param create: bool optional (default False), if True the tag is created if is missing

    Kwargs:
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path

    :returns: xmltree with set text
    """
    from masci_tools.util.xml.xml_setters_xpaths import xml_set_text

    base_xpath = get_tag_xpath(schema_dict, tag_name, **kwargs)

    if complex_xpath is None:
        complex_xpath = base_xpath

    xmltree = xml_set_text(xmltree,
                           schema_dict,
                           complex_xpath,
                           base_xpath,
                           text,
                           occurrences=occurrences,
                           create=create)

    return xmltree


def set_first_text(xmltree, schema_dict, attributename, attribv, complex_xpath=None, create=False, **kwargs):
    """
    Sets the text the first occurrence of a tag in a xmltree to a given value, specified by the name of the tag and
    further specifications. By default the text will be set on all nodes returned for the specified xpath.
    If there are no nodes under the specified xpath a tag can be created with `create=True`.
    The text values are converted automatically according to the types
    with :py:func:`~masci_tools.util.xml.converters.convert_text_to_xml()` if they
    are not `str` already.

    :param xmltree: an xmltree that represents inp.xml
    :param schema_dict: InputSchemaDict containing all information about the structure of the input
    :param tag_name: str name of the tag, where the text should be set
    :param text: value or list of values to set
    :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
    :param create: bool optional (default False), if True the tag is created if is missing

    Kwargs:
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path

    :returns: xmltree with set text
    """
    return set_text(xmltree,
                    schema_dict,
                    attributename,
                    attribv,
                    complex_xpath=complex_xpath,
                    create=create,
                    occurrences=0,
                    **kwargs)


def set_simple_tag(xmltree, schema_dict, tag_name, changes, complex_xpath=None, create_parents=False, **kwargs):
    """
    Sets one or multiple `simple` tag(s) in an xmltree. A simple tag can only hold attributes and has no
    subtags. The tag is specified by its name and further specification
    If the tag can occur multiple times all existing tags are DELETED and new ones are written.
    If the tag only occurs once it will automatically be created if its missing.

    :param xmltree: an xmltree that represents inp.xml
    :param schema_dict: InputSchemaDict containing all information about the structure of the input
    :param tag_name: str name of the tag to modify/set
    :param changes: list of dicts or dict with the changes. Elements in list describe multiple tags.
                    Keys in the dictionary correspond to {'attributename': attributevalue}
    :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
    :param create_parents: bool optional (default False), if True and the path, where the simple tags are
                           set does not exist it is created

    Kwargs:
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path

    :returns: xmltree with set simple tags
    """
    from masci_tools.util.xml.xml_setters_xpaths import xml_set_simple_tag
    from masci_tools.util.xml.common_functions import split_off_tag

    base_xpath = get_tag_xpath(schema_dict, tag_name, **kwargs)

    #Since we can set multiple simple tags we need to provide the path for the parent
    parent_xpath, tag_name = split_off_tag(base_xpath)

    tag_info = schema_dict['tag_info'][base_xpath]

    assert len(tag_info['simple'] | tag_info['complex']) == 0, f"Given tag '{tag_name}' is not simple"

    if complex_xpath is None:
        complex_xpath = parent_xpath

    return xml_set_simple_tag(xmltree,
                              schema_dict,
                              complex_xpath,
                              parent_xpath,
                              tag_name,
                              changes,
                              create_parents=create_parents)


def set_complex_tag(xmltree, schema_dict, tag_name, changes, complex_xpath=None, create=False, **kwargs):
    """
    Function to correctly set tags/attributes for a given tag.
    Goes through the attributedict and decides based on the schema_dict, how the corresponding
    key has to be handled.
    The tag is specified via its name and evtl. further specification

    Supports:

        - attributes
        - tags with text only
        - simple tags, i.e. only attributes (can be optional single/multiple)
        - complex tags, will recursively create/modify them

    :param xmltree: an xmltree that represents inp.xml
    :param schema_dict: InputSchemaDict containing all information about the structure of the input
    :param tag_name: name of the tag to set
    :param attributedict: Keys in the dictionary correspond to names of tags and the values are the modifications
                          to do on this tag (attributename, subdict with changes to the subtag, ...)
    :param complex_xpath: an optional xpath to use instead of the simple xpath for the evaluation
    :param create: bool optional (default False), if True and the path, where the complex tag is
                   set does not exist it is created

    Kwargs:
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path

    :returns: xmltree with changes to the complex tag
    """
    from masci_tools.util.xml.xml_setters_xpaths import xml_set_complex_tag

    base_xpath = get_tag_xpath(schema_dict, tag_name, **kwargs)

    if complex_xpath is None:
        complex_xpath = base_xpath

    return xml_set_complex_tag(xmltree, schema_dict, complex_xpath, base_xpath, changes, create=create)


def set_species_label(xmltree, schema_dict, atom_label, attributedict, create=False):
    """
    This method calls :func:`~masci_tools.util.xml.xml_setters_names.set_species()`
    method for a certain atom species that corresponds to an atom with a given label

    :param xmltree: xml etree of the inp.xml
    :param schema_dict: InputSchemaDict containing all information about the structure of the input
    :param atom_label: string, a label of the atom which specie will be changed. 'all' to change all the species
    :param attributedict: a python dict specifying what you want to change.
    :param create: bool, if species does not exist create it and all subtags?

    :returns: xml etree of the new inp.xml
    """
    from masci_tools.util.schema_dict_util import tag_exists, eval_simple_xpath
    from masci_tools.util.xml.common_functions import get_xml_attribute

    if atom_label == 'all':
        return set_species(xmltree, schema_dict, 'all', attributedict, create=create)

    atom_label = '{: >20}'.format(atom_label)
    all_groups = eval_simple_xpath(xmltree, schema_dict, 'atomGroup', list_return=True)

    species_to_set = set()

    # set all species, where given label is present
    for group in all_groups:
        if tag_exists(group, schema_dict, 'filmPos'):
            atoms = eval_simple_xpath(group, schema_dict, 'filmPos', list_return=True)
        else:
            atoms = eval_simple_xpath(group, schema_dict, 'relPos', list_return=True)
        for atom in atoms:
            label = get_xml_attribute(atom, 'label')
            if label == atom_label:
                species_to_set.add(get_xml_attribute(group, 'species'))

    for species_name in species_to_set:
        xmltree = set_species(xmltree, schema_dict, species_name, attributedict, create=create)

    return xmltree


def set_species(xmltree, schema_dict, species_name, attributedict, create=False):
    """
    Method to set parameters of a species tag of the fleur inp.xml file.

    :param xmltree: xml etree of the inp.xml
    :param schema_dict: InputSchemaDict containing all information about the structure of the input
    :param species_name: string, name of the specie you want to change
                         Can be name of the species, 'all' or 'all-<string>' (sets species with the string in the species name)
    :param attributedict: a python dict specifying what you want to change.
    :param create: bool, if species does not exist create it and all subtags?

    :raises ValueError: if species name is non existent in inp.xml and should not be created.
                        also if other given tags are garbage. (errors from eval_xpath() methods)

    :return xmltree: xml etree of the new inp.xml

    **attributedict** is a python dictionary containing dictionaries that specify attributes
    to be set inside the certain specie. For example, if one wants to set a MT radius it
    can be done via::

        attributedict = {'mtSphere' : {'radius' : 2.2}}

    Another example::

        'attributedict': {'special': {'socscale': 0.0}}

    that switches SOC terms on a sertain specie. ``mtSphere``, ``atomicCutoffs``,
    ``energyParameters``, ``lo``, ``electronConfig``, ``nocoParams``, ``ldaU`` and
    ``special`` keys are supported. To find possible
    keys of the inner dictionary please refer to the FLEUR documentation flapw.de
    """
    from masci_tools.util.xml.xml_setters_xpaths import xml_set_complex_tag

    base_xpath_species = get_tag_xpath(schema_dict, 'species')

    # TODO lowercase everything
    # TODO make a general specifier for species, not only the name i.e. also
    # number, other parameters
    if species_name == 'all':
        xpath_species = base_xpath_species
    elif species_name[:4] == 'all-':  #format all-<string>
        xpath_species = f'{base_xpath_species}[contains(@name,"{species_name[4:]}")]'
    else:
        xpath_species = f'{base_xpath_species}[@name = "{species_name}"]'

    return xml_set_complex_tag(xmltree, schema_dict, xpath_species, base_xpath_species, attributedict, create=create)


def shift_value_species_label(xmltree, schema_dict, atom_label, attributename, value_given, mode='abs', **kwargs):
    """
    Shifts the value of an attribute on a species by label
    if atom_label contains 'all' then applies to all species

    :param xmltree: xml etree of the inp.xml
    :param schema_dict: InputSchemaDict containing all information about the structure of the input
    :param atom_label: string, a label of the atom which specie will be changed. 'all' if set up all species
    :param attributename: name of the attribute to change
    :param value_given: value to add or to multiply by
    :param mode: 'rel' for multiplication or 'abs' for addition

    Kwargs if the attributename does not correspond to a unique path:
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path

    :returns: xml etree of the new inp.xml
    """
    from masci_tools.util.schema_dict_util import tag_exists, eval_simple_xpath
    from masci_tools.util.xml.common_functions import get_xml_attribute
    from masci_tools.util.xml.xml_setters_xpaths import xml_add_number_to_first_attrib
    from masci_tools.util.xml.common_functions import split_off_attrib

    if 'contains' in kwargs:
        contains = kwargs.get('contains')
        if not isinstance(contains, list):
            contains = [contains]
        contains.append('species')
        kwargs['contains'] = contains
    else:
        kwargs['contains'] = 'species'

    species_base_path = get_tag_xpath(schema_dict, 'species')
    attr_base_path = get_attrib_xpath(schema_dict, attributename, **kwargs)
    tag_base_xpath, attributename = split_off_attrib(attr_base_path)

    if atom_label != 'all':
        atom_label = '{: >20}'.format(atom_label)
    all_groups = eval_simple_xpath(xmltree, schema_dict, 'atomGroup', list_return=True)

    species_to_set = set()

    for group in all_groups:
        if tag_exists(group, schema_dict, 'filmPos'):
            atoms = eval_simple_xpath(group, schema_dict, 'filmPos', list_return=True)
        else:
            atoms = eval_simple_xpath(group, schema_dict, 'relPos', list_return=True)
        for atom in atoms:
            label = get_xml_attribute(atom, 'label')
            if atom_label in ('all', label):
                species_to_set.add(get_xml_attribute(group, 'species'))

    for species_name in species_to_set:

        xpath_species = f'{species_base_path}[@name="{species_name}"]'
        tag_xpath = tag_base_xpath.replace(species_base_path, xpath_species)

        xmltree = xml_add_number_to_first_attrib(xmltree,
                                                 schema_dict,
                                                 tag_xpath,
                                                 tag_base_xpath,
                                                 attributename,
                                                 value_given,
                                                 mode=mode)

    return xmltree


def set_atomgroup_label(xmltree, schema_dict, atom_label, attributedict, create=False):
    """
    This method calls :func:`~masci_tools.util.xml.xml_setters_names.set_atomgroup()`
    method for a certain atom species that corresponds to an atom with a given label.

    :param xmltree: xml etree of the inp.xml
    :param schema_dict: InputSchemaDict containing all information about the structure of the input
    :param atom_label: string, a label of the atom which specie will be changed. 'all' to change all the species
    :param attributedict: a python dict specifying what you want to change.
    :param create: bool, if species does not exist create it and all subtags?

    :returns: xml etree of the new inp.xml

    **attributedict** is a python dictionary containing dictionaries that specify attributes
    to be set inside the certain specie. For example, if one wants to set a beta noco parameter it
    can be done via::

        'attributedict': {'nocoParams': {'beta': val}}

    """
    from masci_tools.util.schema_dict_util import tag_exists, eval_simple_xpath
    from masci_tools.util.xml.common_functions import get_xml_attribute

    if atom_label == 'all':
        xmltree = set_atomgroup(xmltree, schema_dict, attributedict, position=None, species='all')
        return xmltree

    atom_label = '{: >20}'.format(atom_label)
    all_groups = eval_simple_xpath(xmltree, schema_dict, 'atomGroup', list_return=True)

    species_to_set = set()

    # set all species, where given label is present
    for group in all_groups:
        if tag_exists(group, schema_dict, 'filmPos'):
            atoms = eval_simple_xpath(group, schema_dict, 'filmPos', list_return=True)
        else:
            atoms = eval_simple_xpath(group, schema_dict, 'relPos', list_return=True)
        for atom in atoms:
            label = get_xml_attribute(atom, 'label')
            if label == atom_label:
                species_to_set.add(get_xml_attribute(group, 'species'))

    for species_name in species_to_set:
        xmltree = set_atomgroup(xmltree, schema_dict, attributedict, position=None, species=species_name)

    return xmltree


def set_atomgroup(xmltree, schema_dict, attributedict, position=None, species=None, create=False):
    """
    Method to set parameters of an atom group of the fleur inp.xml file.

    :param xmltree: xml etree of the inp.xml
    :param schema_dict: InputSchemaDict containing all information about the structure of the input
    :param attributedict: a python dict specifying what you want to change.
    :param position: position of an atom group to be changed. If equals to 'all', all species will be changed
    :param species: atom groups, corresponding to the given species will be changed
    :param create: bool, if species does not exist create it and all subtags?

    :returns: xml etree of the new inp.xml

    **attributedict** is a python dictionary containing dictionaries that specify attributes
    to be set inside the certain specie. For example, if one wants to set a beta noco parameter it
    can be done via::

        'attributedict': {'nocoParams': {'beta': val}}

    """
    from masci_tools.util.xml.xml_setters_xpaths import xml_set_complex_tag

    atomgroup_base_path = get_tag_xpath(schema_dict, 'atomGroup')
    atomgroup_xpath = atomgroup_base_path

    if not position and not species:  # not specfied what to change
        return xmltree

    if position:
        if not position == 'all':
            atomgroup_xpath = f'{atomgroup_base_path}[{position}]'
    if species:
        if not species == 'all':
            atomgroup_xpath = f'{atomgroup_base_path}[@species = "{species}"]'

    xmltree = xml_set_complex_tag(xmltree,
                                  schema_dict,
                                  atomgroup_xpath,
                                  atomgroup_base_path,
                                  attributedict,
                                  create=create)

    return xmltree


def shift_value(xmltree, schema_dict, change_dict, mode='abs', path_spec=None):
    """
    Shifts numerical values of attributes directly in the inp.xml file.

    The first occurrence of the attribute is shifted

    :param xmltree: xml tree that represents inp.xml
    :param schema_dict: InputSchemaDict containing all information about the structure of the input
    :param change_dict: a python dictionary with the keys to shift and the shift values.
    :param mode: 'abs' if change given is absolute, 'rel' if relative
    :param path_spec: dict, with ggf. necessary further specifications for the path of the attribute

    :returns: a xml tree with shifted values

    An example of change_dict::

            change_dict = {'itmax' : 1, 'dVac': -0.123}
    """
    from masci_tools.util.case_insensitive_dict import CaseInsensitiveDict

    if path_spec is None:
        path_spec = {}
    path_spec = CaseInsensitiveDict(path_spec)

    for key, value_given in change_dict.items():

        key_spec = path_spec.get(key, {})
        #This method only support unique and unique_path attributes
        if 'exclude' not in key_spec:
            key_spec['exclude'] = ['other']
        elif 'other' not in key_spec['exclude']:
            key_spec['exclude'].append('other')

        xmltree = add_number_to_first_attrib(xmltree, schema_dict, key, value_given, mode=mode, **key_spec)

    return xmltree


def set_inpchanges(xmltree, schema_dict, change_dict, path_spec=None):
    """
    This method sets all the attribute and texts provided in the change_dict.

    The first occurrence of the attribute/tag is set

    :param xmltree: xml tree that represents inp.xml
    :param schema_dict: InputSchemaDict containing all information about the structure of the input
    :params change_dict: dictionary {attrib_name : value} with all the wanted changes.
    :param path_spec: dict, with ggf. necessary further specifications for the path of the attribute

    An example of change_dict::

            change_dict = {'itmax' : 1,
                           'l_noco': True,
                           'ctail': False,
                           'l_ss': True}

    :returns: an xmltree of the inp.xml file with changes.
    """
    from masci_tools.util.xml.xml_setters_xpaths import xml_set_first_attrib_value, xml_set_first_text
    from masci_tools.util.xml.common_functions import split_off_attrib
    from masci_tools.util.case_insensitive_dict import CaseInsensitiveDict

    if path_spec is None:
        path_spec = {}
    path_spec = CaseInsensitiveDict(path_spec)

    for key, change_value in change_dict.items():

        #Special alias for xcFunctional since name is not a very telling attribute name
        if key == 'xcFunctional':
            key = 'name'

        if key not in schema_dict['attrib_types'] and key not in schema_dict['simple_elements']:
            raise ValueError(f"You try to set the key:'{key}' to : '{change_value}', but the key is unknown"
                             ' to the fleur plug-in')

        text_attrib = key not in schema_dict['attrib_types']

        key_spec = path_spec.get(key, {})
        #This method only support unique and unique_path attributes
        if 'exclude' not in key_spec:
            key_spec['exclude'] = ['other']
        elif 'other' not in key_spec['exclude']:
            key_spec['exclude'].append('other')

        key_xpath = get_attrib_xpath(schema_dict, key, **key_spec)

        if not text_attrib:
            #Split up path into tag path and attribute name (original name of key could have different cases)
            key_xpath, key = split_off_attrib(key_xpath)

        if text_attrib:
            xml_set_first_text(xmltree, schema_dict, key_xpath, key_xpath, change_value)
        else:
            xml_set_first_attrib_value(xmltree, schema_dict, key_xpath, key_xpath, key, change_value)

    return xmltree


@schema_dict_version_dispatch(output_schema=False)
def set_nkpts(xmltree, schema_dict, count, gamma):
    """
    Sets a k-point mesh directly into inp.xml (Only available for inputs before Max4)

    :param xmltree: xml tree that represents inp.xml
    :param schema_dict: InputSchemaDict containing all information about the structure of the input
    :param count: number of k-points
    :param gamma: bool that controls if the gamma-point should be included
                  in the k-point mesh

    :returns: an xmltree of the inp.xml file with changes.
    """

    raise NotImplementedError(f"'set_npkts' is not implemented for inputs of version '{schema_dict['inp_version']}'")


@set_nkpts.register(max_version='0.31')
def set_nkpts_max4(xmltree, schema_dict, count, gamma):
    """
    Sets a k-point mesh directly into inp.xml specific for inputs of version Max4

    :param xmltree: xml tree that represents inp.xml
    :param schema_dict: InputSchemaDict containing all information about the structure of the input
    :param count: number of k-points
    :param gamma: bool that controls if the gamma-point should be included
                  in the k-point mesh

    :returns: an xmltree of the inp.xml file with changes.
    """
    from masci_tools.util.schema_dict_util import eval_simple_xpath, tag_exists

    if not tag_exists(xmltree, schema_dict, 'kPointCount', not_contains='altKPoint'):
        bzintegration_tag = eval_simple_xpath(xmltree, schema_dict, 'bzIntegration')

        for child in bzintegration_tag.iterchildren():
            if 'kPoint' in child.tag:
                bzintegration_tag.remove(child)

        xmltree = create_tag(xmltree, schema_dict, 'kPointCount', not_contains='altKPoint')

    xmltree = set_attrib_value(xmltree, schema_dict, 'count', count, contains='kPointCount', not_contains='altKPoint')
    xmltree = set_attrib_value(xmltree, schema_dict, 'gamma', gamma, contains='kPointCount', not_contains='altKPoint')

    return xmltree



# -*- coding: utf-8 -*-
"""
Test of the functions in masci_tools.util.xml.common_functions
"""
import pytest
import os
from pprint import pprint
import logging

TEST_FOLDER = os.path.dirname(os.path.abspath(__file__))
CLEAR_XML_TEST_FILE = os.path.abspath(os.path.join(TEST_FOLDER, 'files/fleur/test_clear.xml'))

LOGGER = logging.getLogger(__name__)


def test_eval_xpath(caplog):
    """
    Test of the eval_xpath function
    """
    from lxml import etree
    from masci_tools.util.xml.common_functions import eval_xpath

    parser = etree.XMLParser(attribute_defaults=True, encoding='utf-8')
    xmltree = etree.parse(CLEAR_XML_TEST_FILE, parser)
    root = xmltree.getroot()

    scfLoop = eval_xpath(root, '//scfLoop')
    assert isinstance(scfLoop, etree._Element)

    scfLoop = eval_xpath(root, '//scfLoop', list_return=True)
    assert len(scfLoop) == 1
    assert isinstance(scfLoop[0], etree._Element)

    include_tags = eval_xpath(root,
                              '//xi:include',
                              namespaces={'xi': 'http://www.w3.org/2001/XInclude'},
                              list_return=True)
    assert len(include_tags) == 2
    assert isinstance(include_tags[0], etree._Element)

    species_z = eval_xpath(root, "//species[@name='Cu-1']/@atomicNumber")
    assert species_z == '29'

    ldau_tags = eval_xpath(root, "//species[@name='Cu-1']/ldaU")
    assert ldau_tags == []

    with pytest.raises(ValueError, match='There was a XpathEvalError on the xpath:'):
        ldau_tags = eval_xpath(root, "//species/[@name='Cu-1']/ldaU")

    with caplog.at_level(logging.WARNING):
        with pytest.raises(ValueError, match='There was a XpathEvalError on the xpath:'):
            ldau_tags = eval_xpath(root, "//species/[@name='Cu-1']/ldaU", logger=LOGGER)

    assert 'There was a XpathEvalError on the xpath:' in caplog.text


def test_clear_xml():
    """
    Test of the clear_xml function
    """
    from lxml import etree
    from masci_tools.util.xml.common_functions import eval_xpath, clear_xml
    parser = etree.XMLParser(attribute_defaults=True, encoding='utf-8')
    xmltree = etree.parse(CLEAR_XML_TEST_FILE, parser)

    #Check that the file contains comments and includes
    root = xmltree.getroot()
    comments = eval_xpath(root, '//comment()', list_return=True)
    assert len(comments) == 3

    include_tags = eval_xpath(root,
                              '//xi:include',
                              namespaces={'xi': 'http://www.w3.org/2001/XInclude'},
                              list_return=True)
    assert len(include_tags) == 2

    symmetry_tags = eval_xpath(root, '//symOp', list_return=True)
    assert len(symmetry_tags) == 0

    cleared_tree = clear_xml(xmltree)
    cleared_root = cleared_tree.getroot()
    old_root = xmltree.getroot()

    #Make sure that the original tree was not modified
    comments = eval_xpath(old_root, '//comment()', list_return=True)
    assert len(comments) == 3

    #Check that the cleared tree is correct
    comments = eval_xpath(cleared_root, '//comment()', list_return=True)
    assert len(comments) == 0

    include_tags = eval_xpath(cleared_root,
                              '//xi:include',
                              namespaces={'xi': 'http://www.w3.org/2001/XInclude'},
                              list_return=True)
    assert len(include_tags) == 0

    symmetry_tags = eval_xpath(cleared_root, '//symOp', list_return=True)
    assert len(symmetry_tags) == 16


def test_get_xml_attribute(caplog):
    """
    Test of the clear_xml function
    """
    from lxml import etree
    from masci_tools.util.xml.common_functions import get_xml_attribute, eval_xpath
    parser = etree.XMLParser(attribute_defaults=True, encoding='utf-8')
    xmltree = etree.parse(CLEAR_XML_TEST_FILE, parser)
    root = xmltree.getroot()

    scfLoop = eval_xpath(root, '//scfLoop')
    assert get_xml_attribute(scfLoop, 'alpha') == '.05000000'
    assert get_xml_attribute(scfLoop, 'itmax') == '1'
    assert get_xml_attribute(scfLoop, 'maxIterBroyd') == '99'

    with pytest.raises(ValueError, match='Tried to get attribute: "TEST" from element scfLoop.'):
        get_xml_attribute(scfLoop, 'TEST')

    with caplog.at_level(logging.WARNING):
        assert get_xml_attribute(scfLoop, 'TEST', logger=LOGGER) is None
    assert 'Tried to get attribute: "TEST" from element scfLoop.' in caplog.text

    with pytest.raises(TypeError,
                       match='Can not get attributename: "TEST" from node of type <class \'lxml.etree._ElementTree\'>'):
        get_xml_attribute(xmltree, 'TEST')

    with caplog.at_level(logging.WARNING):
        assert get_xml_attribute(xmltree, 'TEST', logger=LOGGER) is None
    assert 'Can not get attributename: "TEST" from node of type <class \'lxml.etree._ElementTree\'>' in caplog.text


def test_split_off_tag():
    """
    Test of the split_off_tag function
    """
    from masci_tools.util.xml.common_functions import split_off_tag

    assert split_off_tag('/fleurInput/calculationSetup/cutoffs') == ('/fleurInput/calculationSetup', 'cutoffs')
    assert split_off_tag('/fleurInput/calculationSetup/cutoffs/') == ('/fleurInput/calculationSetup', 'cutoffs')
    assert split_off_tag('./calculationSetup/cutoffs') == ('./calculationSetup', 'cutoffs')


def test_split_off_attrib():
    """
    Test of the split_off_tag function
    """
    from masci_tools.util.xml.common_functions import split_off_attrib

    assert split_off_attrib('/fleurInput/calculationSetup/cutoffs/@Kmax') == ('/fleurInput/calculationSetup/cutoffs',
                                                                              'Kmax')
    with pytest.raises(AssertionError):
        split_off_attrib('/fleurInput/calculationSetup/cutoffs')
    with pytest.raises(AssertionError):
        split_off_attrib("/fleurInput/atomSpecies/species[@name='TEST']")
    assert split_off_attrib('./calculationSetup/cutoffs/@Kmax') == ('./calculationSetup/cutoffs', 'Kmax')


def test_check_complex_xpath(load_inpxml):
    """
    Test of the check_complex_xpath function
    """
    from masci_tools.util.xml.common_functions import check_complex_xpath

    FILE_PATH = os.path.dirname(os.path.abspath(__file__))
    TEST_INPXML_PATH = os.path.join(FILE_PATH, 'files/fleur/Max-R5/FePt_film_SSFT_LO/files/inp2.xml')

    xmltree, _ = load_inpxml(TEST_INPXML_PATH)

    check_complex_xpath(xmltree, '/fleurInput/atomSpecies/species', "/fleurInput/atomSpecies/species[@name='Fe-1']")

    with pytest.raises(ValueError):
        check_complex_xpath(xmltree, '/fleurInput/atomSpecies/species',
                            "/fleurInput/atomSpecies/species[@name='Fe-1']/lo")

    with pytest.raises(ValueError):
        check_complex_xpath(xmltree, '/fleurInput/atomSpecies/species',
                            "/fleurInput/atomSpecies/species[@name='Fe-1']/@name")

    check_complex_xpath(xmltree, '/fleurInput/atomSpecies/species', '/fleurInput/atomSpecies/species')
    check_complex_xpath(xmltree, '/fleurInput/atomSpecies/species',
                        "/fleurInput/atomSpecies/species[@name='does_not_exist']")
    check_complex_xpath(xmltree, '/fleurInput/atomSpecies/species/lo', "//species[@name='Pt-1']/lo")


def test_abs_to_rel_xpath():

    from masci_tools.util.xml.common_functions import abs_to_rel_xpath

    assert abs_to_rel_xpath('/test/new_root/relative/path', 'new_root') == './relative/path'
    assert abs_to_rel_xpath('/test/new_root/relative/path/@attrib', 'new_root') == './relative/path/@attrib'
    assert abs_to_rel_xpath('/test/new_root/relative/path', 'path') == './'
    assert abs_to_rel_xpath('/test/new_root/relative/path/@attrib', 'path') == './@attrib'

    with pytest.raises(ValueError):
        abs_to_rel_xpath('/test/new_root/relative/path/@attrib', 'non_existent')
